#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 15 08:36:26 2024

@author: mike
"""
import os
import uuid6
import urllib3
import booklet
from typing import Any, Generic, Iterator, Union, List, Dict
import s3func
import weakref
import orjson
import base64
import copy
# import msgspec

# import utils
from . import utils


###############################################
### Parameters

ebooklet_types = ('EVariableLengthValue', 'RemoteConnGroup')


###############################################
### Functions


def get_db_metadata(session, db_key):
    """

    """
    resp_obj = session.head_object(db_key)
    if resp_obj.status == 200:

        meta = resp_obj.metadata
    elif resp_obj.status == 404:
        meta = None
    else:
        raise urllib3.exceptions.HTTPError(resp_obj.error)

    if meta is None:
        return dict(uuid=None, timestamp=None, ebooklet_type=None)
    else:
        meta['uuid'] = uuid6.UUID(hex=meta['uuid'])
        # self.uuid = uuid6.UUID(hex=meta['uuid'])
        # self.ebooklet_type = meta['ebooklet_type']
        meta['timestamp'] = int(meta['timestamp'])

        return meta


def get_user_metadata(session, db_key):
    """

    """
    resp_obj = session.get_object(f'{db_key}/{booklet.utils.metadata_key_bytes.decode()}')
    if resp_obj.status == 200:

        meta = orjson.loads(resp_obj.data)
    elif resp_obj.status == 404:
        meta = None
    else:
        raise urllib3.exceptions.HTTPError(resp_obj.error)

    return meta


def check_write_config(
        access_key_id: str=None,
        access_key: str=None,
        db_key: str=None,
        bucket: str=None,
        endpoint_url: str=None,
        ):
    """

    """
    if isinstance(access_key_id, str) and isinstance(access_key, str) and isinstance(db_key, str) and isinstance(bucket, str):
        if isinstance(endpoint_url, str):
            if not s3func.utils.is_url(endpoint_url):
                raise TypeError(f'{endpoint_url} is not a proper url.')
        return True
    return False


def create_s3_read_session(
        access_key_id: str=None,
        access_key: str=None,
        db_key: str=None,
        bucket: str=None,
        endpoint_url: str=None,
        db_url: str=None,
        threads: int=20,
        read_timeout: int=60,
        retries: int=3,
        ):
    """

    """
    if isinstance(db_url, str):
        if not s3func.utils.is_url(db_url):
            raise TypeError(f'{db_url} is not a proper url.')
        read_session = s3func.HttpSession(threads, read_timeout=read_timeout, stream=False, max_attempts=retries)
        key = db_url
    elif isinstance(access_key_id, str) and isinstance(access_key, str) and isinstance(db_key, str) and isinstance(bucket, str):
        if isinstance(endpoint_url, str):
            if not s3func.utils.is_url(endpoint_url):
                raise TypeError(f'{endpoint_url} is not a proper url.')
        read_session = s3func.S3Session(access_key_id, access_key, bucket, endpoint_url, threads, read_timeout=read_timeout, stream=False, max_attempts=retries)
        key = db_key
    else:
        read_session = None
        # raise ValueError('Either db_url or a combo of access_key_id, access_key, db_key, and bucket (and optionally endpoint_url) must be passed.')

    return read_session, key


def create_s3_write_session(
        access_key_id: str=None,
        access_key: str=None,
        db_key: str=None,
        bucket: str=None,
        endpoint_url: str=None,
        threads: int=20,
        read_timeout: int=60,
        retries: int=3,
        ):
    """

    """
    if check_write_config(access_key_id, access_key, db_key, bucket, endpoint_url):
        write_session = s3func.S3Session(access_key_id, access_key, bucket, endpoint_url, threads, read_timeout=read_timeout, stream=False, max_attempts=retries)
    else:
        write_session = None
        # raise ValueError('Either db_url or a combo of access_key_id, access_key, db_key, and bucket (and optionally endpoint_url) must be passed.')

    return write_session, db_key


# def create_connection(
#         access_key_id: str=None,
#         access_key: str=None,
#         db_key: str=None,
#         bucket: str=None,
#         endpoint_url: str=None,
#         db_url: str=None,
#         threads: int=20,
#         read_timeout: int=60,
#         retries: int=3,
#         meta: dict={},
#         ):
#     """

#     """
#     ## temp read session
#     read_session, key = create_s3_read_session(
#             access_key_id,
#             access_key,
#             db_key,
#             bucket,
#             endpoint_url,
#             db_url,
#             threads,
#             read_timeout,
#             retries,
#             )
#     if read_session is None:
#         raise ValueError('Either db_url or a combo of access_key_id, access_key, db_key, and bucket (and optionally endpoint_url) must be passed.')

#     ## Get metadata if necessary
#     if all([k in meta for k in ('uuid', 'ebooklet_type', 'user_meta')]):
#         uuid = meta['uuid']
#         if isinstance(uuid, (str, uuid6.UUID)):
#             if isinstance(uuid, str):
#                 meta['uuid'] = uuid6.UUID(hex=uuid)
#             else:
#                 raise TypeError('uuid in meta must be either a string or UUID')

#         ebooklet_type = meta['ebooklet_type']
#         if ebooklet_type not in ebooklet_types:
#             raise ValueError(f'ebooklet_type in meta must be one of {ebooklet_types}')
#         if not isinstance(meta['user_meta'], dict):
#             raise TypeError('user_meta in meta must be a dict')
#     else:
#         meta = get_db_metadata(read_session, key)
#     read_session.clear()










###############################################
### Classes

# class ConnParams(msgspec.Struct):
#     db_key: str
#     bucket: str
#     endpoint_url: str
#     type: str
#     uuid: uuid6.UUID


class JsonSerializer:
    def to_dict(self):
        d1 = dict(db_key=self.db_key,
                  bucket=self.bucket,
                  endpoint_url=self.endpoint_url,
                  db_url=self.db_url,
                  # ebooklet_type=self.ebooklet_type,
                  # timestamp=self.timestamp,
                   user_meta=self.user_meta,
                  )
        db_meta = dict(
            ebooklet_type=self.ebooklet_type,
            timestamp=self.timestamp,
            )
        uuid1 = self.uuid
        if uuid1 is not None:
            db_meta['uuid'] = uuid1.hex
        else:
            db_meta['uuid'] = None
        d1['db_meta'] = db_meta

        return d1

    def dumps(self):
        return orjson.dumps(self.to_dict())


class S3SessionReader:
    """

    """
    def __init__(self,
                 read_session,
                 read_db_key,
                 threads,
                 # timestamp,
                 # uuid,
                 # ebooklet_type,
                 # init_bytes,
                 ):
        self._read_session = read_session
        self.read_db_key = read_db_key
        self.threads = threads
        # self.timestamp = timestamp
        # self.uuid = uuid
        # self.ebooklet_type = ebooklet_type
        # self.init_bytes = init_bytes
        self.load_db_metadata()

    def __bool__(self):
        """
        Test to see if remote read access is possible. Same as .read_access.
        """
        return self.read_access

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        """
        Close the remote connection. Should return None.
        """
        if hasattr(self, '_finalizer'):
            self._finalizer()

    def load_db_metadata(self):
        """

        """
        resp_obj = self._read_session.head_object(self.read_db_key)
        if resp_obj.status == 200:
            meta = resp_obj.metadata
            self._init_bytes = base64.urlsafe_b64decode(meta['init_bytes'])
            self.timestamp = int(meta['timestamp'])
            self.uuid = uuid6.UUID(hex=meta['uuid'])
            self.ebooklet_type = meta['ebooklet_type']
        elif resp_obj.status == 404:
            self._init_bytes = None
            self.uuid = None
            self.timestamp = None
            self.ebooklet_type = None
        else:
            raise urllib3.exceptions.HTTPError(resp_obj.error)


    def get_uuid(self):
        """

        """
        if self.uuid is None:
            self.load_metadata()

        return self.uuid


    def get_ebooklet_type(self):
        """

        """
        if self.ebooklet_type is None:
            self.load_metadata()

        return self.ebooklet_type

    def get_timestamp(self):
        """

        """
        resp_obj = self.head_db_object()
        if resp_obj.status == 200:
            self.timestamp = int(resp_obj.metadata['timestamp'])

        elif resp_obj.status == 404:
            self.timestamp = None
        else:
            raise urllib3.exceptions.HTTPError(resp_obj.error)

        return self.timestamp

    # @property
    # def timestamp(self):
    #     """
    #     Timestamp as int_us of the last modified date
    #     """
    #     raise NotImplementedError()

    # @property
    # def readable(self):
    #     """
    #     Test to see if remote read access is possible. Returns a bool.
    #     """
    #     return True

    # @property
    # def writable(self):
    #     """
    #     Test to see if remote write access is possible. Returns a bool.
    #     """
    #     return False

    # def get_db_index_object(self):
    #     """

    #     """
    #     return self._session.get_object(self.db_key + '.remote_index')

    def get_db_object(self):
        """

        """
        return self._read_session.get_object(self.read_db_key)

    # def head_db_object(self):
    #     """

    #     """
    #     resp_obj = self.read_session.head_object(self.db_key)

    #     return resp_obj

    def get_object(self, key: str):
        """
        Get a remote object/file. The input should be a key as a str. It should return an object with a .status attribute as an int, a .data attribute in bytes, and a .error attribute as a dict.
        """
        return self._read_session.get_object(self.read_db_key + '/' + key)

    # def head_object(self, key: str):
    #     """
    #     Get the header for a remote object/file. The input should be a key as a str.
    #     """
    #     return self.read_session.head_object(self.db_key + '/' + key)


class S3SessionWriter(S3SessionReader):
    """

    """
    def __init__(self,
                 read_session,
                 write_session,
                 read_db_key,
                 write_db_key,
                 threads,
                 # timestamp,
                 # uuid,
                 # ebooklet_type,
                 # init_bytes,
                 object_lock=False,
                 break_other_locks=False,
                 lock_timeout=-1
                 ):
        self._read_session = read_session
        self._write_session = write_session
        self.read_db_key = read_db_key
        self.write_db_key = write_db_key
        self.threads = threads
        # self.timestamp = timestamp
        # self.uuid = uuid
        # self.ebooklet_type = ebooklet_type
        # self.init_bytes = init_bytes

        if object_lock:
            lock = write_session.s3lock(self.write_db_key)

            if break_other_locks:
                lock.break_other_locks()

            lock.aquire(timeout=lock_timeout)

            self._writable_check = True
            self._writable = True
        else:
            lock = None

            self._writable_check = False
            self._writable = False

        ## Finalizer
        self._finalizer = weakref.finalize(self, utils.s3session_finalizer, self._write_session, lock)

        ## Get latest metadata
        self.load_db_metadata()

    @property
    def writable(self):
        """
        Should I include this? Or should I simply let the other methods fail if it's not writable? I do like having an explicit test...
        """
        if not self._writable_check:
            test_key = self.write_db_key + uuid6.uuid6().hex[:13]
            put_resp = self._write_session.put_object(test_key, b'0')
            if put_resp.status // 100 == 2:
                del_resp = self._write_session.delete_object(test_key, put_resp.metadata['version_id'])
                if del_resp.status // 100 == 2:
                    self._writable = True

            self._writable_check = True

        return self._writable


    def put_db_object(self, data: bytes, metadata):
        """

        """
        if self.writable:
            return self._write_session.put_object(self.write_db_key, data, metadata=metadata)
        else:
            raise ValueError('Session is not writable.')


    # def put_db_index_object(self, data: bytes, metadata={}):
    #     """

    #     """
    #     if self.writable:
    #         return self._session.put_object(self.db_key + '.remote_index', data, metadata=metadata)
    #     else:
    #         raise ValueError('Conn is not writable.')


    def put_object(self, key: str, data: bytes, metadata={}):
        """

        """
        if self.writable:
            return self._write_session.put_object(self.write_db_key + '/' + key, data, metadata=metadata)
        else:
            raise ValueError('Session is not writable.')


    def delete_objects(self, keys):
        """
        Delete objects
        """
        if self.writable:
            del_list = []
            resp = self._write_session.list_object_versions(prefix=self.write_db_key + '/')
            for obj in resp.iter_objects():
                key0 = obj['key']
                key = key0.split('/')[-1]
                if key in keys:
                    del_list.append({'Key': key0, 'VersionId': obj['version_id']})

            if del_list:
                del_resp = self._write_session.delete_objects(del_list)
                if del_resp.status // 100 != 2:
                    raise urllib3.exceptions.HTTPError(del_resp.error)
        else:
            raise ValueError('Session is not writable.')


    def delete_remote(self):
        """

        """
        if self.writable:
            del_list = []
            resp = self._write_session.list_object_versions(prefix=self.write_db_key)
            for obj in resp.iter_objects():
                key0 = obj['key']
                del_list.append({'Key': key0, 'VersionId': obj['version_id']})

            self._write_session.delete_objects(del_list)
            self._init_bytes = None
            self.uuid = None
        else:
            raise ValueError('Session is not writable.')

    # def rebuild_index(self):
    #     """
    #     Rebuild the remote index file from all the objects in the remote.
    #     """
    #     if self.writable:
    #         resp = self._session.list_object_versions(prefix=self.db_key + '/')


    # def list_objects(self):
    #     """

    #     """
    #     return self.write_session.list_objects(prefix=self.write_db_key)


    # def list_object_versions(self):
    #     """

    #     """
    #     return self.write_session.list_object_versions(prefix=self.write_db_key)

    # def lock(self):
    #     """

    #     """
    #     if self.writable:
    #         lock = self._session.s3lock(self.db_key)
    #         return lock
    #     else:
    #         raise ValueError('Conn is not writable.')


class S3Connection(JsonSerializer):
    """

    """
    def __init__(self,
                access_key_id: str=None,
                access_key: str=None,
                db_key: str=None,
                bucket: str=None,
                endpoint_url: str=None,
                db_url: str=None,
                threads: int=20,
                read_timeout: int=60,
                retries: int=3,
                db_meta: dict=None,
                user_meta: dict=None,
                ):
        """

        """
        ## temp read session
        read_session, key = create_s3_read_session(
                access_key_id,
                access_key,
                db_key,
                bucket,
                endpoint_url,
                db_url,
                threads,
                read_timeout,
                retries,
                )
        if read_session is None:
            raise ValueError('Either db_url or a combo of access_key_id, access_key, db_key, and bucket (and optionally endpoint_url) must be passed.')

        ## Get metadata if necessary
        if db_meta is None:
            meta = get_db_metadata(read_session, key)
        elif all([k in db_meta for k in ('uuid', 'ebooklet_type', 'timestamp')]):
            uuid = db_meta['uuid']
            if isinstance(uuid, (str, uuid6.UUID)):
                if isinstance(uuid, str):
                    db_meta['uuid'] = uuid6.UUID(hex=uuid)
                else:
                    raise TypeError('uuid in meta must be either a string or UUID')

            ebooklet_type = db_meta['ebooklet_type']
            if ebooklet_type not in ebooklet_types:
                raise ValueError(f'ebooklet_type in meta must be one of {ebooklet_types}')
        else:
            meta = get_db_metadata(read_session, key)

        if user_meta is not None:
            if not isinstance(user_meta, dict):
                raise TypeError('user_meta in meta must be a dict or None')

        # read_session.clear()

        ## Assign properties
        self.db_key = db_key
        self.bucket = bucket
        self.access_key_id = access_key_id
        self.access_key = access_key
        self.endpoint_url = endpoint_url
        self.db_url = db_url
        self.threads = threads
        self.read_timeout = read_timeout
        self.retries = retries
        # self.type = 'connection'

        self.uuid = meta['uuid']
        self.timestamp = meta['timestamp']
        self.ebooklet_type = meta['ebooklet_type']
        self.user_meta = user_meta


    def _make_read_session(self):
        """

        """
        read_session, key = create_s3_read_session(
                self.access_key_id,
                self.access_key,
                self.db_key,
                self.bucket,
                self.endpoint_url,
                self.db_url,
                self.threads,
                self.read_timeout,
                self.retries,
                )

        return read_session, key


    def load_db_metadata(self):
        """

        """
        read_session, key = self._make_read_session()

        meta = get_db_metadata(read_session, key)
        self.uuid = meta['uuid']
        self.ebooklet_type = meta['ebooklet_type']
        self.timestamp = meta['timestamp']


    def load_user_metadata(self):
        """

        """
        read_session, key = self._make_read_session()

        self.user_meta = get_user_metadata(read_session, key)


    def open(self,
             flag: str='r',
             object_lock: bool=False,
             break_other_locks: bool=False,
             lock_timeout: int=-1,
             ):
        """

        """
        if (flag != 'r') and (self.access_key_id is None or self.access_key is None):
            raise ValueError("access_key_id and access_key must be assigned to open a connection for writing.")

        # ## Check and load uuid if it isn't assigned
        # if self.uuid is None:
        #     self.load_db_metadata()

        ## Read session
        read_session, read_db_key = self._make_read_session()

        if flag == 'r':
            return S3SessionReader(read_session, read_db_key, self.threads)
        else:
            write_session, write_db_key = create_s3_write_session(
                    self.access_key_id,
                    self.access_key,
                    self.db_key,
                    self.bucket,
                    self.endpoint_url,
                    self.threads,
                    self.read_timeout,
                    self.retries,
                    )

            if write_session is None:
                raise ValueError("A write session could not be created. Check that all the inputs are assigned.")

            session_writer = S3SessionWriter(
                                    read_session,
                                    write_session,
                                    read_db_key,
                                    write_db_key,
                                    self.threads,
                                    object_lock,
                                    break_other_locks,
                                    lock_timeout,
                                    )

            # Check to make sure the uuids are the same if the read and write sessions are different
            if isinstance(read_session, s3func.HttpSession) and self.uuid is not None:
                if self.uuid != session_writer.uuid:
                    raise ValueError('The UUIDs of the http connection and the S3 connection are different. Check to make sure the they are pointing to the right file.')

            return session_writer























































#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""
import io
import os
from collections.abc import Mapping, MutableMapping
from typing import Any, Generic, Iterator, Union, List, Dict
import pathlib
import concurrent.futures
import multiprocessing
import threading
import booklet
import s3func
# import zstandard as zstd
import orjson
import pprint
import tempfile
import weakref
import shutil
import uuid6 as uuid
import urllib3
from itertools import count
from collections import deque
# import portalocker

# import utils
from . import utils

# import remote
from . import remote

# uuid_s3dbm = b'K=d:\xa89F(\xbc\xf5 \xd7$\xbd;\xf2'
# version = 1
# version_bytes = version.to_bytes(2, 'little', signed=False)

# lock_remote = False
# break_other_locks = False

#######################################################
### Classes


class Change:
    """

    """
    def __init__(self, ebooklet):
        """

        """
        self._ebooklet = ebooklet

        self.update()


    def pull(self):
        """

        """
        self._ebooklet.sync()

        ## update the remote timestamp
        self._ebooklet._remote_session.get_timestamp()

        ## Determine if a change has occurred
        overwrite_remote_index = utils.check_local_remote_sync(self._ebooklet._local_file, self._ebooklet._remote_session, self._ebooklet._flag)

        ## Pull down the remote index
        if overwrite_remote_index:
            utils.get_remote_index_file(self._ebooklet._local_file_path, overwrite_remote_index, self._ebooklet._remote_session, self._ebooklet._flag)


    def update(self):
        """

        """
        self._ebooklet.sync()
        changelog_path = utils.create_changelog(self._ebooklet._local_file_path, self._ebooklet._local_file, self._ebooklet._remote_index, self._ebooklet._remote_session)

        self._changelog_path = changelog_path


    def iter_changes(self):
        if not self._changelog_path:
            self.update()
        return utils.view_changelog(self._changelog_path)


    def discard(self, keys=None):
        """
        Removes changed keys in the local file. If keys is None, then removes all changed keys.
        """
        if not self._ebooklet.writable:
            raise ValueError('File is open for read-only.')

        with booklet.FixedValue(self._changelog_path) as f:
            if keys is None:
                rm_keys = f.keys()
            else:
                rm_keys = [key for key in keys if key in f]

            for key in rm_keys:
                # print(key)
                del self._ebooklet._local_file[key]

        self._changelog_path.unlink()
        self._changelog_path = None


    def push(self, force_push=False):
        """
        Updates the remote. It will regenerate the changelog to ensure the changelog is up-to-date. Returns True if the remote has been updated and False if no updates were made (due to nothing needing updating).
        Force_push will push the main file and the remote_index to the remote regardless of changes. Only necessary if upload failures occurred during a previous push.
        """
        if not self._ebooklet._remote_session.writable:
            raise ValueError('Remote is not writable.')

        if not self._ebooklet.writable:
            raise ValueError('File is open for read-only.')

        self.update()

        # if self._remote_index is None:
        #     remote_index = booklet.FixedValue(self._remote_index_path, 'n', key_serializer='str', value_len=7, n_buckets=self._local_file._n_buckets, buffer_size=self._local_file._write_buffer_size)

        #     self._remote_index = remote_index
        #     self._finalizer.detach()
        #     self._finalizer = weakref.finalize(self, utils.ebooklet_finalizer, self._local_file, self._remote_index)

        success = utils.update_remote(self._ebooklet._local_file, self._ebooklet._remote_index, self._changelog_path, self._ebooklet._remote_session, self._ebooklet._executor, force_push, self._ebooklet._deletes, self._ebooklet._flag, self._ebooklet._subtype)

        if success:
            self._changelog_path.unlink()
            self._changelog_path = None # Force a reset of the changelog
            self._ebooklet._deletes.clear()

            if self._ebooklet._remote_session.uuid is None:
                self._ebooklet._remote_session.load_db_metadata()

        return success


# class UserMetadata(MutableMapping):
#     """

#     """
#     def __init__(self, bookcase, book_hash: str=None):
#         """

#         """
#         if isinstance(book_hash, str):
#             user_meta = bookcase._meta['books'][book_hash]['user_metadata']
#         else:
#             user_meta = bookcase._meta['user_metadata']

#         self._bookcase = bookcase
#         self._user_meta = user_meta
#         self._book_hash = book_hash
#         self._modified = False
#         # self._local_meta_path = local_meta_path
#         # self._remote_s3_access = remote_s3_access

#     def __repr__(self):
#         """

#         """
#         return pprint.pformat(self._user_meta)

#     def __setitem__(self, key, value):
#         """

#         """
#         self._user_meta[key] = value
#         self._modified = True


#     def __getitem__(self, key: str):
#         """

#         """
#         return self._user_meta[key]

#     def __delitem__(self, key):
#         """

#         """
#         del self._user_meta[key]
#         self._modified = True

#     def clear(self):
#         """

#         """
#         self._user_meta.clear()
#         self._modified = True


#     def keys(self):
#         """

#         """
#         return self._user_meta.keys()


#     def items(self):
#         """

#         """
#         return self._user_meta.items()


#     def values(self, keys: List[str]=None):
#         return self._user_meta.values()


#     def __iter__(self):
#         return self._user_meta.keys()

#     def __len__(self):
#         """
#         """
#         return len(self._user_meta)


#     def __contains__(self, key):
#         return key in self._user_meta


#     def get(self, key, default=None):
#         return self._user_meta.get(key)


#     def update(self, key_value_dict: Union[Dict[str, bytes], Dict[str, io.IOBase]]):
#         """

#         """
#         self._user_meta.update(key_value_dict)
#         self._modified = True

#     def __enter__(self):
#         return self

#     def __exit__(self, *args):
#         self.close()

#     def close(self):
#         self.sync()

#     def sync(self):
#         """

#         """
#         if self._modified:
#             int_us = utils.make_timestamp()

#             if isinstance(self._book_hash, str):
#                 if not self._bookcase.remote_s3_access:
#                     self._bookcase._meta['books'][self._book_hash]['last_modified'] += 1
#                 else:
#                     self._bookcase._meta['books'][self._book_hash]['last_modified'] = int_us
#                 # self._bookcase._meta['books'][self._book_hash]['last_modified'] = int_us
#                 self._bookcase._meta['books'][self._book_hash]['user_metadata'] = self._user_meta
#             else:
#                 if not self._bookcase.remote_s3_access:
#                     self._bookcase._meta['last_modified'] += 1
#                 else:
#                     self._bookcase._meta['last_modified'] = int_us

#                 self._bookcase._meta['user_metadata'] = self._user_meta

#             utils.write_metadata(self._bookcase._local_meta_path, self._bookcase._meta)




    # def sync(self):
    #     """

    #     """
    #     if self._modified:
    #         int_us = utils.make_timestamp()
    #         self._metadata['last_modified'] = int_us
    #         if self._version_date:
    #             self._metadata['versions'][self._version_position] = {'versions_date': self._version_date, 'user_metadata': self._user_meta}
    #         else:
    #             self._metadata['user_metadata'] = self._user_meta

    #         with io.open(self._local_meta_path, 'wb') as f:
    #             f.write(zstd.compress(orjson.dumps(self._metadata, option=orjson.OPT_NON_STR_KEYS | orjson.OPT_SERIALIZE_NUMPY)))





# class Bookcase:
#     """

#     """
#     def __init__(self,
#                  file_path: Union[str, pathlib.Path],
#                  flag: str = "r",
#                  value_serializer: str = None,
#                  n_buckets: int=12007,
#                  buffer_size: int = 2**22,
#                  remote: Union[remotes.BaseRemote, str]=None
#                  ):
#         """

#         """
#         ## Pre-processing
#         if file_path is None:
#             temp_path = pathlib.Path(tempfile.TemporaryDirectory().name)
#             local_meta_path = temp_path.joinpath('temp.bcs')
#             self._finalizer = weakref.finalize(self, shutil.rmtree, temp_path, True)
#         else:
#             local_meta_path = pathlib.Path(file_path)
#             temp_path = None

#         # local_meta_path = pathlib.Path(local_db_path)
#         remote_keys_name = local_meta_path.name + '.remote_keys'
#         remote_keys_path = local_meta_path.parent.joinpath(remote_keys_name)

#         for key, value in local_storage_kwargs.items():
#             if key not in utils.local_storage_options:
#                 raise ValueError(f'{key} in local_storage_kwargs, but it must only contain {utils.local_storage_options}.')
#         if 'n_buckets' not in local_storage_kwargs:
#             n_buckets = utils.default_n_buckets
#             local_storage_kwargs['n_buckets'] = n_buckets
#         else:
#             n_buckets = int(local_storage_kwargs['n_buckets'])
#         local_storage_kwargs.update({'key_serializer': 'str', 'value_serializer': 'bytes'})
#         if value_serializer in booklet.serializers.serial_name_dict:
#             value_serializer_code = booklet.serializers.serial_name_dict[value_serializer]
#         else:
#             raise ValueError(f'value_serializer must be one of {booklet.available_serializers}.')

#         ## Create S3 lock for writes
#         if write and remote_s3_access:
#             lock = s3func.s3.S3Lock(connection_config, bucket, remote_db_key, read_timeout=read_timeout)
#             if break_other_locks:
#                 lock.break_other_locks()
#             if not lock.aquire(timeout=lock_timeout):
#                 raise TimeoutError('S3Lock timed out')
#         else:
#             lock = None

#         ## Finalizer
#         self._finalizer = weakref.finalize(self, utils.bookcase_finalizer, temp_path, lock)

#         ## Init metadata
#         meta, meta_in_remote = utils.init_metadata(local_meta_path, remote_keys_path, write, http_session, s3_session, remote_s3_access, remote_http_access, remote_url, remote_db_key, value_serializer, local_storage_kwargs)

#         ## Init local storage
#         # local_data_path = utils.init_local_storage(local_meta_path, flag, meta)

#         ## Assign properties
#         # self._temp_path = temp_path
#         self._meta_in_remote = meta_in_remote
#         self._remote_db_key = remote_db_key
#         self._n_buckets = n_buckets
#         self._write = write
#         self._buffer_size = buffer_size
#         self._connection_config = connection_config
#         self._read_timeout = read_timeout
#         self._lock = lock
#         self.remote_s3_access = remote_s3_access
#         self.remote_http_access = remote_http_access
#         self._bucket = bucket
#         self._meta = meta
#         self._threads = threads
#         self._local_meta_path = local_meta_path
#         self._remote_keys_path = remote_keys_path
#         self._value_serializer = value_serializer
#         self._value_serializer_code = value_serializer_code
#         self._local_storage_kwargs = local_storage_kwargs
#         self._host_url = host_url
#         self._remote_base_url = remote_base_url
#         self._remote_url = remote_url
#         self._s3_session = s3_session
#         self._http_session = http_session

#         ## Assign the metadata object for global
#         self.metadata = UserMetadata(self)


#     @property
#     def default_book_name(self):
#         """

#         """
#         if self._meta['default_book']:
#             return self._meta['books'][self._meta['default_book']]['name']

#     def list_book_names(self):
#         """

#         """
#         for key, val in self._meta['books'].items():
#             yield val['name']


#     def set_default_book_name(self, book_name):
#         """

#         """
#         book_hash = utils.hash_book_name(book_name)
#         if book_hash in self._meta['books']:
#             self._meta['default_book'] = book_hash
#             # meta_bytes = zstd.compress(orjson.dumps(self._meta, option=orjson.OPT_NON_STR_KEYS | orjson.OPT_SERIALIZE_NUMPY), level=1)
#             # with io.open(self._local_meta_path, 'wb') as f:
#             #     f.write(meta_bytes)
#         else:
#             raise KeyError(book_name)


#     # def create_book(self, book_name):
#     #     """
#     #     Remove
#     #     """
#     #     meta = utils.create_book(self._local_meta_path, self._meta, book_name, self.remote_s3_access)
#     #     self._meta = meta

#     #     return True


#     def open_book(self, book_name: str=None, flag: str='r'):
#         """
#         Remove the create_book method and include a flag parameter.
#         """
#         if book_name is None:
#             if flag == 'r':
#                 book_hash = self._meta['default_book']
#                 if book_hash is None:
#                     raise KeyError('No books exist. Open with write permissions to create a book.')
#             else:
#                 raise KeyError('book_name must be specified when open for writing.')
#         else:
#             book_hash = utils.hash_book_name(book_name)

#         if flag in ('n', 'c'):
#             if flag == 'c':
#                 if book_hash in self._meta['books']:
#                     raise KeyError(f'{book_name} already exists as a book.')
#             meta = utils.create_book(self._local_meta_path, self._meta, book_name, book_hash, self.remote_s3_access)
#             self._meta = meta

#         book = Book(self, book_hash)

#         return book


#     def close(self):
#         """

#         """
#         if self._flag != 'r':
#             self.metadata.close()

#         self._finalizer()

#     def __enter__(self):
#         return self

#     def __exit__(self, *args):
#         self.close()

#     def pull_remote_index(self, book_name):
#         """

#         """
#         ## Get base path for the book
#         if book_name is None:
#             book_hash = self._meta['default_book']
#         else:
#             book_hash = utils.hash_book_name(book_name)
#             if book_hash not in self._meta['books']:
#                 raise KeyError(book_name)

#         book_file_name = self._local_meta_path.name + f'.{book_hash}.'
#         book_base_path = self._local_meta_path.parent.joinpath(book_file_name)

#         remote_index_path = utils.get_remote_index_file(book_base_path, book_hash, self._remote_db_key, self._remote_url, self._http_session, self._s3_session, self.remote_http_access, self.remote_s3_access, True)

#         return remote_index_path


#     def pull_metadata(self):
#         """

#         """
#         if self._meta_in_remote:
#             self.metadata.sync()
#             meta, meta_in_remote = utils.init_metadata(self._local_meta_path, self._remote_keys_path, self._flag, self._write, self._http_session, self._s3_session, self._remote_s3_access, self._remote_http_access, self._remote_url, self._remote_db_key, self._value_serializer, self._local_storage_kwargs)
#             return True
#         else:
#             return False


class EVariableLengthValue(MutableMapping):
    """

    """
    def __init__(
            self,
            remote_conn: remote.S3Connection,
            file_path: Union[str, pathlib.Path],
            flag: str = "r",
            value_serializer: str = None,
            n_buckets: int=12007,
            buffer_size: int = 2**22,
            object_lock: bool=False,
            break_other_locks: bool=False,
            lock_timeout: int=-1,
            # inherit_remote: Union[remotes.BaseRemote, str]=None,
            # inherit_data: bool=False,
            ):
        """

        """
        ## Inherit another remote
        # if (inherit_remote is not None) and (flag in ('c', 'n')):
        #     if isinstance(inherit_remote, str):
        #         inherit_remote = remotes.HttpRemote(inherit_remote)
        #     elif not isinstance(inherit_remote, remotes.BaseRemote):
        #         raise TypeError('inherit_remote must be either a Remote or a url string.')

            # TODO: Pull down the remote ebooklet and assign it to this new object

        # check_timestamp = init_check_remote

        local_file_path = pathlib.Path(file_path)

        local_file_exists = local_file_path.exists()

        ## Determine the remotes that read and write
        remote_session = utils.check_parse_conn(remote_conn, flag, object_lock, break_other_locks, lock_timeout, local_file_exists)

        ## Init the local file
        local_file, overwrite_remote_index = utils.init_local_file(local_file_path, flag, remote_session, value_serializer, n_buckets, buffer_size)

        remote_index_path = utils.get_remote_index_file(local_file_path, overwrite_remote_index, remote_session, flag)

        # Open remote index file
        if remote_index_path.exists():
            # remote_index = booklet.FixedValue(remote_index_path, 'r')
            if flag == 'r':
                remote_index = booklet.FixedLengthValue(remote_index_path, 'r')
            else:
                remote_index = booklet.FixedLengthValue(remote_index_path, 'w')
        else:
            remote_index = booklet.FixedLengthValue(remote_index_path, 'n', key_serializer='str', value_len=7, n_buckets=n_buckets, buffer_size=buffer_size)

        ## Finalizer
        self._finalizer = weakref.finalize(self, utils.ebooklet_finalizer, local_file, remote_index, remote_session)

        ## Assign properties
        if flag == 'r':
            self.writable = False
        else:
            self.writable = True

        self._flag = flag
        self._remote_index_path = remote_index_path
        self._local_file_path = local_file_path
        self._local_file = local_file
        self._remote_index_path = remote_index_path
        self._remote_index = remote_index
        self._deletes = set()
        # self._read_conn_open = read_conn_open
        # self._write_conn_open = write_conn_open
        self._remote_session = remote_session
        # self._changelog_path = None
        self._n_buckets = local_file._n_buckets
        # self._clear = False
        # self._lock = lock
        self._subtype = 'EVariableLengthValue'
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=remote_session.threads)


    # def _pre_value(self, value) -> bytes:

    #     ## Serialize to bytes
    #     try:
    #         value = self._value_serializer.dumps(value)
    #     except Exception as error:
    #         raise error

    #     return value

    # def _post_value(self, value: bytes):

    #     ## Serialize from bytes
    #     value = self._value_serializer.loads(value)

    #     return value

    def set_metadata(self, data, timestamp=None):
        """
        Sets the metadata for the booklet. The data input must be a json serializable object. Optionally assign a timestamp.
        """
        if self.writable:
            self._local_file.set_metadata(data, timestamp)
        else:
            raise ValueError('File is open for read only.')


    def get_metadata(self, include_timestamp=False):
        """
        Get the metadata. Optionally include the timestamp in the output.
        Will return None if no metadata has been assigned.
        """
        _ = self._load_item(booklet.utils.metadata_key_bytes.decode())
        # if failure:
        #     return failure

        return self._local_file.get_metadata(include_timestamp=include_timestamp)


    def keys(self):
        """

        """
        overlap = set()
        for key in self._local_file.keys():
            if key in self._remote_index:
                overlap.add(key)
            yield key

        for key in self._remote_index.keys():
            if key not in overlap:
                yield key


    def items(self):
        """

        """
        _ = self.load_items()

        return self._local_file.items()

    def values(self):
        _ = self.load_items()

        return self._local_file.values()

    def timestamps(self, include_value=False):
        """
        Return an iterator for timestamps for all keys. Optionally add values to the iterator.
        """
        _ = self.load_items()

        return self._local_file.timestamps(include_value=include_value)


    def get_timestamp(self, key, include_value=False, default=None):
        """
        Get a timestamp associated with a key. Optionally include the value.
        """
        failure = self._load_item(key)
        if failure:
            return failure

        return self._local_file.get_timestamp(key, include_value=include_value, default=default)

    def set_timestamp(self, key, timestamp):
        """
        Set a timestamp for a specific key. The timestamp must be either an int of the number of microseconds in POSIX UTC time, an ISO 8601 datetime string with timezone, or a datetime object with timezone.
        """
        if self.writable:
            self._local_file.set_timestamp(key, timestamp)
        else:
            raise ValueError('File is open for read only.')


    def set(self, key, value, timestamp=None):
        """

        """
        if self.writable:
            self._local_file.set(key, value, timestamp)

            # if self._read_conn.uuid:
            #     int_us = utils.make_timestamp()
            # else:
            #     old_val = self._local_index.get(key)
            #     if old_val:
            #         int_us = utils.bytes_to_int(old_val) + 1
            #     else:
            #         int_us = 0
            # val_bytes = self._pre_value(value)
            # self._local_data[key] = val_bytes
            # self._local_index[key] = utils.int_to_bytes(int_us, 7)
        else:
            raise ValueError('File is open for read only.')



    def __iter__(self):
        return self.keys()

    def __len__(self):
        """

        """
        counter = count()
        deque(zip(self.keys(), counter), maxlen=0)

        return next(counter)


    def __contains__(self, key):
        if (key in self._remote_index) or (key in self._local_file):
            return True
        else:
            return False

    def get(self, key, default=None):
        failure = self._load_item(key)
        if failure:
            return failure

        return self._local_file.get(key, default=default)


    def update(self, key_value_dict: dict):
        """

        """
        if self.writable:
            for key, value in key_value_dict.items():
                self[key] = value
        else:
            raise ValueError('File is open for read only.')


    def prune(self, timestamp=None, reindex=False):
        """
        Prunes the old keys and associated values. Returns the number of removed items. The method can also prune remove keys/values older than the timestamp. The user can also reindex the booklet file. False does no reindexing, True increases the n_buckets to a preassigned value, or an int of the n_buckets. True can only be used if the default n_buckets were used at original initialisation.
        """
        if self.writable:
            removed = self._local_file.prune(timestamp=timestamp, reindex=reindex)
            self._n_buckets = self._local_file._n_buckets

            _ = self._remote_index.prune(reindex)

            return removed
        else:
            raise ValueError('File is open for read only.')


    def get_items(self, keys, default=None):
        """

        """
        _ = self.load_items(keys)

        for key in keys:
            output = self._local_file.get(key, default=default)
            if output is None:
                yield key, None
            else:
                # ts, value = output
                yield key, output


    def load_items(self, keys=None):
        """
        Loads items into the local file without returning the values. If keys is None, then it loads all of the values in the remote. Returns a dict of keys with the errors trying to access the remote.
        """
        futures = {}
        failure_dict = {}

        writable = self._local_file.writable

        if not writable:
            self._local_file.reopen('w')

        if keys is None:
            for key, remote_time_bytes in self._remote_index.items():
                check = utils.check_local_vs_remote(self._local_file, remote_time_bytes, key)
                if check:
                    f = self._executor.submit(utils.get_remote_value, self._local_file, key, self._remote_session)
                    futures[f] = key
        else:
            for key in keys:
                remote_time_bytes = self._remote_index.get(key)
                check = utils.check_local_vs_remote(self._local_file, remote_time_bytes, key)
                if check:
                    f = self._executor.submit(utils.get_remote_value, self._local_file, key, self._remote_session)
                    futures[f] = key

        for f in concurrent.futures.as_completed(futures):
            key = futures[f]
            error = f.result()
            if error is not None:
                failure_dict[key] = error

        if not writable:
            self._local_file.reopen('r')

        return failure_dict


    def _load_item(self, key):
        """

        """
        # if key in self._deletes:
        #     raise KeyError(key)

        remote_time_bytes = self._remote_index.get(key)
        check = utils.check_local_vs_remote(self._local_file, remote_time_bytes, key)

        if check:
            if not self._local_file.writable:
                self._local_file.reopen('w')
                failure = utils.get_remote_value(self._local_file, key, self._remote_session)
                self._local_file.reopen('r')
            else:
                failure = utils.get_remote_value(self._local_file, key, self._remote_session)
            return failure
        else:
            return None


    def __getitem__(self, key: str):
        value = self.get(key)

        if value is None:
            raise KeyError(f'{key}')
        else:
            return value


    def __setitem__(self, key: str, value):
        if self.writable:
            self.set(key, value)
        else:
            raise ValueError('File is open for read only.')

    def __delitem__(self, key):
        if self.writable:
            if key in self._remote_index:
                del self._remote_index[key]
                self._deletes.add(key)

            if key in self._local_file:
                del self._local_file[key]
        else:
            raise ValueError('File is open for read only.')

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def clear(self, local_only=True):
        if self.writable:
            self._local_file.clear()

            # if not local_only:
            #     if self._remote_index is not None:
            #         self._remote_index.clear()
        else:
            raise ValueError('File is open for read only.')

    def close(self, force_close=False):
        self.sync()
        self._executor.shutdown(cancel_futures=force_close)
        # self._manager.shutdown()
        self._finalizer()


    # def __del__(self):
    #     self.close()

    def sync(self):
        self._executor.shutdown()
        del self._executor
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=self._remote_session.threads)
        self._remote_index.sync()
        self._local_file.sync()

    def changes(self):
        return Change(self)


    def reopen(self, flag):
        """

        """
        self.close()
        self._local_file.reopen(flag)
        self._remote_index.reopen(flag)

        # if self._lock is not None:
        #     self._lock.aquire()

        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=self._remote_session.threads)

        self._finalizer = weakref.finalize(self, utils.ebooklet_finalizer, self._local_file, self._remote_index, self._remote_session)


    # def pull(self):
    #     """

    #     """
    #     self.sync()
    #     self._read_conn._parse_db_object()
    #     overwrite_remote_index = utils.check_local_remote_sync(self._local_file, self._read_conn)
    #     if overwrite_remote_index:
    #         utils.get_remote_index_file(self._local_file_path, overwrite_remote_index, self._read_conn)


    # def update_changelog(self):
    #     """

    #     """
    #     self.sync()
    #     changelog_path = utils.create_changelog(self._local_file_path, self._local_file, self._remote_index, self._read_conn)

    #     self._changelog_path = changelog_path


    # def changelog(self):
    #     if not self._changelog_path:
    #         self.update_changelog()
    #     return utils.view_changelog(self._changelog_path)


    # def push(self, force_push=False):
    #     """
    #     Updates the remote. It will regenerate the changelog to ensure the changelog is up-to-date. Returns True if the remote has been updated and False if no updates were made (due to nothing needing updating).
    #     Force_push will push the main file and the remote_index to the remote regardless of changes. Only necessary if upload failures occurred during a previous push.
    #     """
    #     if not self._write_conn.writable:
    #         raise ValueError('Remote is not writable.')

    #     self.update_changelog()

    #     # if self._remote_index is None:
    #     #     remote_index = booklet.FixedValue(self._remote_index_path, 'n', key_serializer='str', value_len=7, n_buckets=self._local_file._n_buckets, buffer_size=self._local_file._write_buffer_size)

    #     #     self._remote_index = remote_index
    #     #     self._finalizer.detach()
    #     #     self._finalizer = weakref.finalize(self, utils.ebooklet_finalizer, self._local_file, self._remote_index)

    #     success = utils.update_remote(self._local_file_path, self._remote_index_path, self._local_file, self._remote_index, self._changelog_path, self._write_conn, self._executor, force_push, self._deletes, self._flag)

    #     if success:
    #         self._changelog_path.unlink()
    #         self._changelog_path = None # Force a reset of the changelog

    #         if self._read_conn.uuid is None:
    #             self._read_conn._parse_db_object()
    #             self._write_conn._parse_db_object()

    #     return success


class RemoteConnGroup(EVariableLengthValue):
    """

    """
    def __init__(
            self,
            remote_conn: remote.S3Connection,
            file_path: Union[str, pathlib.Path],
            flag: str = "r",
            n_buckets: int=12007,
            buffer_size: int = 2**22,
            object_lock: bool=False,
            break_other_locks: bool=False,
            lock_timeout: int=-1,
            # inherit_remote: Union[remotes.BaseRemote, str]=None,
            # inherit_data: bool=False,
            ):
        """

        """
        ## Inherit another remote
        # if (inherit_remote is not None) and (flag in ('c', 'n')):
        #     if isinstance(inherit_remote, str):
        #         inherit_remote = remotes.HttpRemote(inherit_remote)
        #     elif not isinstance(inherit_remote, remotes.BaseRemote):
        #         raise TypeError('inherit_remote must be either a Remote or a url string.')

            # TODO: Pull down the remote ebooklet and assign it to this new object

        # check_timestamp = init_check_remote

        local_file_path = pathlib.Path(file_path)

        local_file_exists = local_file_path.exists()

        ## Determine the remotes that read and write
        remote_session = utils.check_parse_conn(remote_conn, flag, object_lock, break_other_locks, lock_timeout, local_file_exists)

        ## Init the local file
        local_file, overwrite_remote_index = utils.init_local_file(local_file_path, flag, remote_session, 'orjson', n_buckets, buffer_size)

        remote_index_path = utils.get_remote_index_file(local_file_path, overwrite_remote_index, remote_session, flag)

        # Open remote index file
        if remote_index_path.exists():
            # remote_index = booklet.FixedValue(remote_index_path, 'r')
            if flag == 'r':
                remote_index = booklet.FixedLengthValue(remote_index_path, 'r')
            else:
                remote_index = booklet.FixedLengthValue(remote_index_path, 'w')
        else:
            remote_index = booklet.FixedLengthValue(remote_index_path, 'n', key_serializer='str', value_len=7, n_buckets=n_buckets, buffer_size=buffer_size)

        ## Finalizer
        self._finalizer = weakref.finalize(self, utils.ebooklet_finalizer, local_file, remote_index, remote_session)

        ## Assign properties
        if flag == 'r':
            self.writable = False
        else:
            self.writable = True

        self._flag = flag
        self._remote_index_path = remote_index_path
        self._local_file_path = local_file_path
        self._local_file = local_file
        self._remote_index_path = remote_index_path
        self._remote_index = remote_index
        self._deletes = set()
        # self._read_conn_open = read_conn_open
        # self._write_conn_open = write_conn_open
        self._remote_session = remote_session
        # self._changelog_path = None
        self._n_buckets = local_file._n_buckets
        # self._clear = False
        # self._lock = lock
        self._subtype = 'RemoteConnGroup'
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=remote_session.threads)


    def add(self, remote_conn: remote.S3Connection):
        """

        """
        if self.writable:

            if not isinstance(remote_conn, remote.S3Connection):
                raise TypeError('remote_conn/value must be a remote.S3Connection')

            ## Update remote_conn meta
            remote_conn.load_db_metadata()
            remote_conn.load_user_metadata()

            uuid0 = remote_conn.uuid
            if uuid0 is None:
                raise ValueError('Remote does not exist. It must exist to be added to a RemoteConnGroup.')

            uuid_hex = uuid0.hex

            value = remote_conn.to_dict()

            self._local_file.set(uuid_hex, value, remote_conn.timestamp)

        else:
            raise ValueError('File is open for read only.')


    def set(self, key, remote_conn: remote.S3Connection):
        """

        """
        if self.writable:

            if not isinstance(remote_conn, remote.S3Connection):
                raise TypeError('remote_conn/value must be a remote.S3Connection')

            ## Update remote_conn meta
            remote_conn.load_db_metadata()
            remote_conn.load_user_metadata()

            uuid0 = remote_conn.uuid
            if uuid0 is None:
                raise ValueError('Remote does not exist. It must exist to be added to a RemoteConnGroup.')

            uuid_hex = uuid0.hex

            if key != uuid_hex:
                raise KeyError('The key must be the uuid hex.')

            value = remote_conn.to_dict()

            self._local_file.set(uuid_hex, value, remote_conn.timestamp)

        else:
            raise ValueError('File is open for read only.')



def open(
    file_path: Union[str, pathlib.Path],
    flag: str = "r",
    value_serializer: str = None,
    n_buckets: int=12007,
    buffer_size: int = 2**22,
    remote_conn: Union[remote.S3Connection, str]=None,
    ebooklet_type: str='EVariableLengthValue',
    ):
    """
    Open an S3 dbm-style database. This allows the user to interact with an S3 bucket like a MutableMapping (python dict) object. If remote_conn is not passed, then it opens a normal booklet file.

    Parameters
    -----------
    file_path : str or pathlib.Path
        It must be a path to a local file location. If you want to use a tempfile, then use the name from the NamedTemporaryFile initialized class.

    flag : str
        Flag associated with how the file is opened according to the dbm style. See below for details.

    key_serializer : str, class, or None
        The serializer to use to convert the input value to bytes. Run the booklet.available_serializers to determine the internal serializers that are available. None will require bytes as input. A custom serializer class can also be used. If the objects can be serialized to json, then use orjson or msgpack. They are super fast and you won't have the pickle issues.

    value_serializer : str, class, or None
        Similar to the key_serializer, except for the values.

    n_buckets : int
        The number of hash buckets to using in the indexing. Generally use the same number of buckets as you expect for the total number of keys.

    buffer_size : int
        The buffer memory size in bytes used for writing. Writes are first written to a block of memory, then once the buffer if filled up it writes to disk. This is to reduce the number of writes to disk and consequently the CPU write overhead.
        This is only used when the file is open for writing.

    remote_conn : S3Connection, str, or None
        The object to connect to a remote. It can either be a Conn type object, an http url string, or None. If None, no remote connection is made and the file is only opened locally.
    
    ebooklet_type : str
        What type of ebooklet to create. Options are either EVariableLengthValue (default) or RemoteConnGroup.

    Returns
    -------
    Ebooklet or Booklet

    The optional *flag* argument can be:

    +---------+-------------------------------------------+
    | Value   | Meaning                                   |
    +=========+===========================================+
    | ``'r'`` | Open existing database for reading only   |
    |         | (default)                                 |
    +---------+-------------------------------------------+
    | ``'w'`` | Open existing database for reading and    |
    |         | writing                                   |
    +---------+-------------------------------------------+
    | ``'c'`` | Open database for reading and writing,    |
    |         | creating it if it doesn't exist           |
    +---------+-------------------------------------------+
    | ``'n'`` | Always create a new, empty database, open |
    |         | for reading and writing                   |
    +---------+-------------------------------------------+

    """
    if remote_conn is None:
        return booklet.open(file_path, flag=flag, key_serializer='str', value_serializer=value_serializer, n_buckets=n_buckets, buffer_size=buffer_size)
    else:
        if isinstance(remote_conn, str):
            if flag != 'r':
                raise ValueError('If remote_conn is a url string, then flag must be r.')
            remote_conn = remote.S3Connection(db_url=remote_conn)
        elif not isinstance(remote_conn, remote.S3Connection):
            raise TypeError('remote_conn must be either a url string or aremote.S3Connection.')

        if remote_conn.uuid is not None:
            if isinstance(remote_conn.ebooklet_type, str):
                if remote_conn.ebooklet_type == 'EVariableLengthValue':
                    return EVariableLengthValue(remote_conn=remote_conn, file_path=file_path, flag=flag, value_serializer=value_serializer, n_buckets=n_buckets, buffer_size=buffer_size)
                elif remote_conn.ebooklet_type == 'RemoteConnGroup':
                    return RemoteConnGroup(remote_conn=remote_conn, file_path=file_path, flag=flag, n_buckets=n_buckets, buffer_size=buffer_size)
                else:
                    raise ValueError('What kind of EBooklet is this?!')
        else:
            if ebooklet_type == 'EVariableLengthValue':
                return EVariableLengthValue(remote_conn=remote_conn, file_path=file_path, flag=flag, value_serializer=value_serializer, n_buckets=n_buckets, buffer_size=buffer_size)
            elif ebooklet_type == 'RemoteConnGroup':
                return RemoteConnGroup(remote_conn=remote_conn, file_path=file_path, flag=flag, n_buckets=n_buckets, buffer_size=buffer_size)
            else:
                raise ValueError('ebooklet_type must be either EVariableLengthValue or RemoteConnGroup.')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan  5 11:04:13 2023

@author: mike
"""
import os
import io
# from pydantic import BaseModel, HttpUrl
import pathlib
import copy
# from time import sleep
# import hashlib
# from hashlib import blake2b
import booklet
import orjson
# from s3func import S3Session, HttpSession
import urllib3
import shutil
from datetime import datetime, timezone
import base64
# import zstandard as zstd
# from glob import glob
import portalocker
import concurrent.futures
# from collections.abc import Mapping, MutableMapping
# from __init__ import __version__ as version

# import remote
from . import remote

############################################
### Parameters

# version = '0.1.0'

default_n_buckets = 100003

# blt_files = ('.local_data', '.remote_index')

# local_storage_options = ('write_buffer_size', 'n_bytes_file', 'n_bytes_key', 'n_bytes_value', 'n_buckets')

int_to_bytes = booklet.utils.int_to_bytes
bytes_to_int = booklet.utils.bytes_to_int

############################################
### Exception classes


# class BaseError(Exception):
#     def __init__(self, message, objs=[], temp_path=None, *args):
#         self.message = message # without this you may get DeprecationWarning
#         # Special attribute you desire with your Error,
#         # for file in blt_files:
#         #     f = getattr(obj, file)
#         #     if f is not None:
#         #         f.close()
#         for obj in objs:
#             if obj:
#                 obj.close()
#         if temp_path:
#             temp_path.cleanup()
#         # allow users initialize misc. arguments as any other builtin Error
#         super(BaseError, self).__init__(message, *args)


# class S3dbmValueError(BaseError):
#     pass

# class S3dbmTypeError(BaseError):
#     pass

# class S3dbmKeyError(BaseError):
#     pass

# class S3dbmHttpError(BaseError):
#     pass

# class S3dbmSerializeError(BaseError):
#     pass


############################################
### Functions


def fake_finalizer():
    """
    The finalizer function for S3Remote instances.
    """


def s3session_finalizer(session, lock):
    """
    The finalizer function for S3Remote instances.
    """
    session.client.close()
    if lock is not None:
        lock.release()


# def bookcase_finalizer(temp_path, lock):
#     """
#     The finalizer function for bookcase instances.
#     """
#     if temp_path:
#         shutil.rmtree(temp_path, True)
#     if lock:
#         lock.release()


def ebooklet_finalizer(local_file, remote_index, remote_session):
    """
    The finalizer function for book instances.
    """
    local_file.close()
    # if remote_index is not None:
    remote_index.close()
    remote_session.close()
    # if read_conn is not None:
    #     read_conn.close()
    # if write_conn is not None:
    #     write_conn.close()
    # if lock is not None:
    #     lock.release()


# def write_metadata(local_meta_path, meta):
#     """

#     """
#     meta_bytes = orjson.dumps(meta, option=orjson.OPT_SERIALIZE_NUMPY)
#     with io.open(local_meta_path, 'wb') as f:
#         f.write(meta_bytes)


# def get_save_remote_file(local_path, remote_url, remote_db_key, http_session, s3_session, remote_s3_access, remote_http_access):
#     """

#     """
#     if remote_http_access:
#         func = http_session.get_object
#         key = remote_url
#     else:
#         func = s3_session.get_object
#         key = remote_db_key

#     resp = func(key)
#     if resp.status == 200:
#         with open(local_path, 'wb') as f:
#             f.write(resp.data)
#     else:
#         raise urllib3.exceptions.HTTPError(resp.error)


# def get_remote_file(remote_url, remote_db_key, http_session, s3_session, remote_s3_access, remote_http_access):
#     """

#     """
#     if remote_http_access:
#         func = http_session.get_object
#         key = remote_url
#     else:
#         func = s3_session.get_object
#         key = remote_db_key

#     resp = func(key)
#     if resp.status == 200:
#         return resp.data
#     else:
#         raise urllib3.exceptions.HTTPError(resp.error)


def check_parse_conn(remote_conn, flag, object_lock, break_other_locks, lock_timeout, local_file_exists):
    """

    """
    if object_lock and (flag != 'r'):
        remote_session = remote_conn.open(flag, object_lock, break_other_locks, lock_timeout)
    else:
        remote_session = remote_conn.open(flag)

    # if isinstance(remote_conn, str):
    #     if flag != 'r':
    #         raise ValueError('If remote_conn is a url string, then flag must be r.')
    #     remote_conn = remote.S3Connection(db_url=remote_conn)
    #     remote_session = remote_conn.open('r')

    # elif isinstance(remote_conn, remote.S3Connection):
    #     if object_lock and (flag != 'r'):
    #         remote_session = remote_conn.open(flag, object_lock, break_other_locks, lock_timeout)
    #     else:
    #         remote_session = remote_conn.open(flag)

    # else:
    #     raise TypeError('The remote_conn must be either an S3Connection object or a url string.')

    if flag in ('r', 'w') and (remote_session.uuid is None) and not local_file_exists:
        raise ValueError('No file was found in the remote, but the local file was open for read and write without creating a new file.')
    # elif flag != 'r' and remote_session is None and not local_file_exists:
    #     raise ValueError('If open for write, then an S3Remote object must be passed.')

    return remote_session


def check_local_remote_sync(local_file, remote_session, flag):
    """

    """
    overwrite_remote_index = False

    remote_uuid = remote_session.uuid
    if remote_uuid and flag != 'n':
        local_uuid = local_file.uuid

        if remote_uuid != local_uuid:
            raise ValueError('The local file has a different UUID than the remote. Use a different local file path or delete the existing one.')

        ## Check timestamp to determine if the local remote_index needs to be updated
        if (remote_session.timestamp > local_file._file_timestamp):
            overwrite_remote_index = True

    return overwrite_remote_index


def init_local_file(local_file_path, flag, remote_session, value_serializer, n_buckets, buffer_size):
    """

    """
    remote_uuid = remote_session.uuid

    if local_file_path.exists():

        if flag == 'n':
            local_file = booklet.open(local_file_path, flag='n', key_serializer='str', value_serializer=value_serializer, n_buckets=n_buckets, buffer_size=buffer_size)

            overwrite_remote_index = True
        else:
            if flag == 'r':
                local_file = booklet.open(local_file_path, 'r')
            else:
                local_file = booklet.open(local_file_path, 'w')

            overwrite_remote_index = check_local_remote_sync(local_file, remote_session, flag)

    else:
        if remote_uuid:
            ## Init with the remote bytes - keeps the remote uuid and timestamp
            local_file = booklet.open(local_file_path, flag='n', init_bytes=remote_session._init_bytes)
            local_file._n_keys = 0

            overwrite_remote_index = True
        else:
            ## Else create a new file
            local_file = booklet.open(local_file_path, flag='n', key_serializer='str', value_serializer=value_serializer, n_buckets=n_buckets, buffer_size=buffer_size)

            overwrite_remote_index = True

    return local_file, overwrite_remote_index


def get_remote_index_file(local_file_path, overwrite_remote_index, remote_session, flag):
    """

    """
    remote_index_path = local_file_path.parent.joinpath(local_file_path.name + '.remote_index')

    if (not remote_index_path.exists() or overwrite_remote_index) and (flag != 'n'):
        if remote_session.uuid:
            # remote_index_key = read_conn.db_key + '.remote_index'

            index0 = remote_session.get_db_object()
            if index0.status == 200:
                with portalocker.Lock(remote_index_path, 'wb', timeout=120) as f:
                    f.write(index0.data)
                    # shutil.copyfileobj(index0.data, f)
            elif index0.status != 404:
                raise urllib3.exceptions.HTTPError(index0.error)

    return remote_index_path


# def init_remote_config(bucket, connection_config, remote_url, threads, read_timeout, retries):
#     """

#     """
#     http_session = None
#     s3_session = None
#     remote_s3_access = False
#     remote_http_access = False
#     remote_base_url = None
#     host_url = None

#     if remote_url is not None:
#         url_grp = urllib3.util.parse_url(remote_url)
#         if url_grp.scheme is not None:
#             http_session = HttpSession(threads, read_timeout=read_timeout, stream=False, max_attempts=retries)
#             url_path = pathlib.Path(url_grp.path)
#             remote_base_url = url_path.parent
#             host_url = url_grp.scheme + '://' + url_grp.host
#             remote_http_access = True
#         else:
#             print(f'{remote_url} is not a proper url.')
#     if (bucket is not None) and (connection_config is not None):
#         s3_session = S3Session(connection_config, bucket, threads, read_timeout=read_timeout, stream=False, max_attempts=retries)
#         remote_s3_access = True

#     # if (not remote_s3_access) and (flag != 'r'):
#     #     raise ValueError("If flag != 'r', then the appropriate remote write access parameters must be passed.")

#     return http_session, s3_session, remote_s3_access, remote_http_access, host_url, remote_base_url


# def init_metadata(local_meta_path, remote_keys_path, write, http_session, s3_session, remote_s3_access, remote_http_access, remote_url, remote_db_key, value_serializer, local_storage_kwargs):
#     """

#     """
#     meta_in_remote = False
#     # get_remote_keys = False
#     remote_meta = None

#     if remote_s3_access:
#         int_us = make_timestamp()
#     else:
#         int_us = 0

#     new_meta = {
#         'package_version': version,
#         'local_storage_kwargs': local_storage_kwargs,
#         'value_serializer': value_serializer,
#         'last_modified': int_us,
#         'user_metadata': {},
#         'default_book': None
#         }

#     glob_str = str(local_meta_path) + '.*'
#     extra_files = glob(glob_str)

#     if local_meta_path.exists():
#         # if write:
#         #     portalocker.lock(local_meta_path, portalocker.LOCK_EX)
#         # else:
#         #     portalocker.lock(local_meta_path, portalocker.LOCK_SH)
#         with io.open(local_meta_path, 'rb') as f:
#             local_meta = orjson.loads(f.read())
#     elif not write:
#         raise FileExistsError('Bookcase was open for read-only, but nothing exists to open.')
#     else:
#         for file in extra_files:
#             os.remove(file)
#         local_meta = None

#     if remote_http_access or remote_s3_access:
#         if remote_http_access:
#             func = http_session.get_object
#             key = remote_url
#         else:
#             func = s3_session.get_object
#             key = remote_db_key

#         meta0 = func(key)
#         if meta0.status == 200:
#             if meta0.metadata['file_type'] != 'bookcase':
#                 raise TypeError('The remote db file is not a bookcase file.')
#             remote_meta = orjson.loads(meta0.data)
#             meta_in_remote = True

#             ## Remote meta is the only meta needed now
#             meta = remote_meta

#             ## Determine if the local remote index files needs to be removed
#             if local_meta is not None:
#                 remote_ts = remote_meta['last_modified']
#                 local_ts = local_meta['last_modified']
#                 if remote_ts > local_ts:
#                     # get_remote_keys_file(local_meta_path, remote_db_key, remote_url, http_session, s3_session, remote_http_access)

#                     # with open(local_meta_path, 'wb') as f:
#                     #     f.write(meta0.data)

#                     local_books = local_meta['books'].copy()
#                     remote_books = remote_meta['books']

#                     # local_remote_index_ts = {}
#                     # for file in extra_files:
#                     #     if '.remote_index' in file:
#                     #         book_hash = file.split('.')[-2]
#                     #         local_book_ts = books[book_hash]['last_modified']
#                     #         local_remote_index_ts[book_hash] = local_book_ts

#                     for file in extra_files:
#                         if '.remote_index' in file:
#                             book_hash = file.split('.')[-2]
#                             local_book_ts = local_books[book_hash]['last_modified']
#                             remote_book = remote_books.get(book_hash)
#                             if remote_book:
#                                 remote_book_ts = remote_book['last_modified']
#                                 if remote_book_ts > local_book_ts:
#                                     os.remove(file)

#                     ## Combining the local and remote books ensures a complete list if local was opened and changed offline
#                     local_books.update(remote_books)
#                     meta['books'] = local_books

#         elif meta0.status == 404:
#             if local_meta is None:
#                 meta = new_meta
#             else:
#                 meta = local_meta
#         else:
#             raise urllib3.exceptions.HTTPError(meta0.error)

#     elif local_meta:
#         meta = local_meta
#     else:
#         meta = new_meta

#     return meta, meta_in_remote


# def create_book(local_meta_path, meta, book_name, book_hash, remote_s3_access):
#     """

#     """
#     if remote_s3_access:
#         int_us = make_timestamp()
#     else:
#         int_us = 0

#     meta['books'][book_hash] = {'last_modified': int_us, 'name': book_name, 'user_metadata': {}}
#     if meta['default_book'] is None:
#         meta['default_book'] = book_hash

#     meta['last_modified'] = int_us

#     # write_metadata(local_meta_path, meta)
#     return meta








# def get_remote_index_file(book_base_path, book_hash, remote_db_key, remote_url, http_session, s3_session, remote_http_access, remote_s3_access, overwrite=False):
#     """

#     """
#     remote_index_path = book_base_path.parent.joinpath(book_base_path.name + 'remote_index')

#     if not remote_index_path.exists() or overwrite:
#         if remote_http_access:
#             remote_index_key = remote_url + f'{book_hash}.remote_index'
#             func = http_session.get_object
#         elif remote_s3_access:
#             remote_index_key = remote_db_key + f'{book_hash}.remote_index'
#             func = s3_session.get_object
#         else:
#             return remote_index_path

#         index0 = func(remote_index_key)
#         if index0.status == 200:
#             with open(remote_index_path, 'wb') as f:
#                 shutil.copyfileobj(index0.data, f)
#         elif index0.status != 404:
#             raise urllib3.exceptions.HTTPError(index0.error)

#     return remote_index_path


def get_remote_value(local_file, key, remote_session):
    """

    """
    resp = remote_session.get_object(key)

    if resp.status == 200:
        timestamp = int(resp.metadata['timestamp'])

        # print(timestamp)
        # print(resp.data)

        # val_bytes = resp.data

        local_file.set(key, resp.data, timestamp, encode_value=False)
    # elif resp.status == 404:
    #     raise KeyError(f'{key} not found in remote.')
    else:
        return resp.error
        # return urllib3.exceptions.HttpError(f'{key} returned the http error {resp.status}.')

    return None


def check_local_vs_remote(local_file, remote_time_bytes, key):
    """

    """
    # remote_time_bytes = remote_index.get(key)

    if remote_time_bytes is None:
        return None

    remote_time_int = bytes_to_int(remote_time_bytes)
    local_time_int = local_file.get_timestamp(key)

    if local_time_int:
        if remote_time_int <= local_time_int:
            return False

    return True


# def check_local_vs_remote_iter(local_file, remote_index, keys):
#     """

#     """
#     for key in keys:
#         yield check_local_vs_remote(local_file, remote_index, key)

# def get_value(local_file, remote_index, key, read_conn):
#     """

#     """
#     if remote_index is not None:
#         # if key not in remote_index:
#         #     return False

#         remote_time_bytes = remote_index.get(key)

#         if remote_time_bytes is None:
#             return local_file.get_timestamp(key, include_value=True)

#         remote_time_int = bytes_to_int(remote_time_bytes)
#         local_time_int_val_bytes = local_file.get_timestamp(key, include_value=True, decode_value=False)

#         if local_time_int_val_bytes:
#             timestamp, val_bytes = local_time_int_val_bytes
#             if remote_time_int <= timestamp:
#                 return timestamp, local_file._post_value(val_bytes)

#         timestamp, val_bytes = get_remote_value(local_file, key, read_conn)

#         return timestamp, local_file._post_value(val_bytes)

#     else:
#         return local_file.get_timestamp(key, include_value=True)


# def load_value(local_file, remote_index, key, read_conn):
#     """

#     """
#     if check_local_vs_remote(local_file, remote_index, key):
#         get_remote_value(local_file, key, read_conn)

#     else:
#         return key in local_file


# def get_value(local_data, local_index, remote_index, key, bucket=None, s3_client=None, session=None, host_url=None, remote_base_url=None):
#     """

#     """
#     if key in local_data:
#         if local_index:
#             local_time_bytes = local_index[key]
#         else:
#             return local_data[key]
#     else:
#         local_time_bytes = None

#     if remote_index:
#         if key not in remote_index:
#             return None

#         remote_time_bytes = remote_index[key]

#         if local_time_bytes:
#             remote_mod_time_int = bytes_to_int(remote_time_bytes)
#             local_mod_time_int = bytes_to_int(local_time_bytes)
#             if remote_mod_time_int < local_mod_time_int:
#                 return local_data[key]

#         value_bytes = get_remote_value(local_data, local_index, remote_index, key, bucket, s3_client, session, host_url, remote_base_url)

#     else:
#         value_bytes = None

#     return value_bytes


#################################################
### local/remote changelog


# def create_changelog(local_data_path, remote_index_path, local_meta_path, n_buckets, meta_in_remote):
#     """
#     Only check and save by the microsecond timestamp. Might need to add in the md5 hash if this is not sufficient.
#     """
#     changelog_path = local_meta_path.parent.joinpath(local_meta_path.name + '.changelog')
#     with booklet.FixedValue(changelog_path, 'n', key_serializer='str', value_len=14, n_buckets=n_buckets) as f:
#         with booklet.VariableValue(local_data_path) as local_data:
#             if meta_in_remote:
#                 # shutil.copyfile(remote_keys_path, temp_remote_keys_path)
#                 # f = booklet.FixedValue(temp_remote_keys_path, 'w')
#                 with booklet.VariableValue(remote_index_path) as remote_index:
#                     for key, local_bytes_us in local_data.items():
#                         remote_val = remote_index.get(key)
#                         if remote_val:
#                             local_int_us = bytes_to_int(local_bytes_us)
#                             remote_bytes_us = remote_val[:7]
#                             remote_int_us = bytes_to_int(remote_bytes_us)
#                             if local_int_us > remote_int_us:
#                                 f[key] = local_bytes_us + remote_bytes_us
#                         else:
#                             f[key] = local_bytes_us + int_to_bytes(0, 7)
#             else:
#                 # f = booklet.FixedValue(temp_remote_keys_path, 'n', key_serializer='str', value_len=26)
#                 for key, local_bytes_us in local_data.items():
#                     f[key] = local_bytes_us + int_to_bytes(0, 7)

#     return changelog_path


def create_changelog(local_file_path, local_file, remote_index, remote_session):
    """
    Only check and save by the microsecond timestamp. Might need to add in the md5 hash if this is not sufficient.
    """
    changelog_path = local_file_path.parent.joinpath(local_file_path.name + '.changelog')
    if remote_index is not None:
        n_buckets = remote_index._n_buckets
    else:
        n_buckets = local_file._n_buckets

    with booklet.FixedLengthValue(changelog_path, 'n', key_serializer='str', value_len=14, n_buckets=n_buckets) as f:
        if remote_session.uuid and remote_index is not None:
            for key, local_int_us in local_file.timestamps():
                remote_bytes_us = remote_index.get(key)
                if remote_bytes_us:
                    remote_int_us = bytes_to_int(remote_bytes_us)
                    if local_int_us > remote_int_us:
                        local_bytes_us = int_to_bytes(local_int_us, 7)
                        f[key] = local_bytes_us + remote_bytes_us
                else:
                    local_bytes_us = int_to_bytes(local_int_us, 7)
                    f[key] = local_bytes_us + int_to_bytes(0, 7)
        else:
            for key, local_int_us in local_file.timestamps():
                local_bytes_us = int_to_bytes(local_int_us, 7)
                f[key] = local_bytes_us + int_to_bytes(0, 7)

    return changelog_path


def view_changelog(changelog_path):
    """

    """
    with booklet.FixedLengthValue(changelog_path) as f:
        for key, val in f.items():
            local_bytes_us = val[:7]
            remote_bytes_us = val[7:]
            local_int_us = bytes_to_int(local_bytes_us)
            remote_int_us = bytes_to_int(remote_bytes_us)
            if remote_int_us == 0:
                remote_ts = None
            else:
                remote_ts = datetime.fromtimestamp(remote_int_us*0.000001, tz=timezone.utc)

            dict1 = {
                'key': key,
                'remote_timestamp': remote_ts,
                'local_timestamp': datetime.fromtimestamp(local_int_us*0.000001, tz=timezone.utc)
                }

            yield dict1


# def check_changelog_keys(


##############################################
### Update remote


def update_remote(local_file, remote_index, changelog_path, remote_session, executor, force_push, deletes, flag, ebooklet_type):
    """

    """
    ## Make sure the files are synced
    # local_file.sync()
    # remote_index.sync()

    ## If file was open for replacement (n), then delete everything in the remote
    if flag == 'n':
        remote_session.delete_remote()

    ## Upload data and update the remote_index file
    # remote_index.reopen('w')

    futures = {}
    with booklet.FixedLengthValue(changelog_path) as cl:
        for key in cl:
            # remote_key = base_remote_key + '/' + key
            time_int_us, valb = local_file.get_timestamp(key, include_value=True, decode_value=False)
            f = executor.submit(remote_session.put_object, key, valb, {'timestamp': str(time_int_us)})
            futures[f] = key

    ## Check the uploads to see if any fail
        updated = False
        failures = []
        for future in concurrent.futures.as_completed(futures):
            key = futures[future]
            run_result = future.result()
            if run_result.status // 100 == 2:
                remote_index[key] = cl[key][:7]
                updated = True
            else:
                failures.append(key)

        if failures:
            print(f"These items failed to upload: {', '.join(failures)}")

    ## Upload the remote_index file
    remote_index.sync()

    if updated or force_push or deletes:
        time_int_us = booklet.utils.make_timestamp_int()

        ## Get main file init bytes
        local_file._set_file_timestamp(time_int_us)
        local_file._file.seek(0)
        local_init_bytes = bytearray(local_file._file.read(200))
        if local_init_bytes[:16] != booklet.utils.uuid_variable_blt:
            raise ValueError(local_init_bytes)

        n_keys_pos = booklet.utils.n_keys_pos
        local_init_bytes[n_keys_pos:n_keys_pos+4] = b'\x00\x00\x00\x00'

        # remote_index_key = write_conn_open.db_key + '.remote_index'
        remote_index._file.seek(0)

        # futures = {}
        # f = executor.submit(write_conn_open.put_db_object, remote_index._file.read(), {'timestamp': str(time_int_us), 'uuid': remote_index.uuid.hex, 'ebooklet_type': ebooklet_type, 'init_bytes': base64.b64encode(local_init_bytes)})
        # futures[f] = write_conn_open.db_key

        resp = remote_session.put_db_object(remote_index._file.read(), {'timestamp': str(time_int_us), 'uuid': remote_index.uuid.hex, 'ebooklet_type': ebooklet_type, 'init_bytes': base64.urlsafe_b64encode(local_init_bytes).decode()})

        # remote_index.reopen('r')

        if resp.status // 100 != 2:
            urllib3.exceptions.HTTPError("The db object failed to upload. You need to rerun the push with force_push=True or the remote will be corrupted.")


        # ## remove deletes in remote
        # if deletes:
        #     write_conn_open.delete_objects(deletes)
        #     for key in deletes:
        #         del remote_index[key]
        #     remote_index.sync()

        ## Save main file init bytes
        # local_file._set_file_timestamp(time_int_us)
        # local_file._file.seek(0)
        # local_init_bytes = bytearray(local_file._file.read(200))
        # if local_init_bytes[:16] != booklet.utils.uuid_variable_blt:
        #     raise ValueError(local_init_bytes)

        # n_keys_pos = booklet.utils.n_keys_pos
        # local_init_bytes[n_keys_pos:n_keys_pos+4] = b'\x00\x00\x00\x00'

        # ## Upload the init bytes
        # f = executor.submit(write_conn_open.put_db_object, local_init_bytes, {'timestamp': str(time_int_us), 'uuid': local_file.uuid.hex})
        # futures[f] = write_conn_open.db_key

        # failures = []
        # for future in concurrent.futures.as_completed(futures):
        #     key = futures[future]
        #     run_result = future.result()
        #     if run_result.status // 100 == 2:
        #         failures.append(key)

        # if failures:
        #     urllib3.exceptions.HTTPError(f"These items failed to upload: {', '.join(failures)}. You need to rerun the push with force_push=True or the remote will be corrupted.")

        ## remove deletes in remote
        if deletes:
            remote_session.delete_objects(deletes)
            deletes.clear()

        return True
    else:
        # remote_index.reopen('r')
        return False















































# def attach_prefix(prefix, key):
#     """

#     """
#     if key == '':
#         new_key = prefix
#     elif not prefix.startswith('/'):
#         new_key = prefix + '/' + prefix


# def test_path(path: pathlib.Path):
#     """

#     """
#     return path


def determine_file_obj_size(file_obj):
    """

    """
    pos = file_obj.tell()
    size = file_obj.seek(0, io.SEEK_END)
    file_obj.seek(pos)

    return size


# def check_local_storage_kwargs(local_storage, local_storage_kwargs, local_file_path):
#     """

#     """
#     if local_storage == 'blt':
#         if 'flag' in local_storage_kwargs:
#             if local_storage_kwargs['flag'] not in ('w', 'c', 'n'):
#                 local_storage_kwargs['flag'] = 'c'
#         else:
#             local_storage_kwargs['flag'] = 'c'

#         local_storage_kwargs['file_path'] = local_file_path

#     return local_storage_kwargs



























































